import typer

from rich import print as rprint


def register_alloy_composition(app: typer.Typer):
    from tc.alloy.types import Alloy

    @app.command(name="composition")
    def alloy_composition(alloy: Alloy) -> None:
        """List known alloy composition."""
        from tc.alloy.known_alloy import get_known_alloy_composition

        try:
            composition = get_known_alloy_composition(alloy)
            print(composition)
        except:
            rprint("⚠️  [yellow]Unable to find composition for alloy[/yellow]")
            _ = typer.Exit()

    _ = alloy_composition



from pydantic import BaseModel
from pathlib import Path


class Composition(BaseModel):
    name: str

    # Base elements
    Fe: float | None = None
    C: float | None = None

    # TCFe alloys
    H: float | None = None
    Mg: float | None = None
    Ca: float | None = None
    Y: float | None = None
    Ti: float | None = None
    Zr: float | None = None
    V: float | None = None
    Nb: float | None = None
    Ta: float | None = None
    Cr: float | None = None
    Mo: float | None = None
    W: float | None = None
    Mn: float | None = None
    Ru: float | None = None
    Co: float | None = None
    Ni: float | None = None
    Cu: float | None = None
    Zn: float | None = None
    B: float | None = None
    Al: float | None = None
    Si: float | None = None
    N: float | None = None
    P: float | None = None
    O: float | None = None
    S: float | None = None
    Ar: float | None = None
    Ce: float | None = None

    # TCNi12 extra elements
    Hf: float | None = None
    Re: float | None = None
    Pd: float | None = None
    Pt: float | None = None

    # TCAl9 extra elements
    Li: float | None = None
    Na: float | None = None
    K: float | None = None
    Be: float | None = None
    Sr: float | None = None
    Ba: float | None = None
    Sc: float | None = None
    Ga: float | None = None
    Ge: float | None = None
    Ag: float | None = None
    In: float | None = None
    Cd: float | None = None
    Sn: float | None = None
    Sb: float | None = None
    Te: float | None = None
    Se: float | None = None
    Pb: float | None = None
    Bi: float | None = None
    La: float | None = None
    Pr: float | None = None
    Nd: float | None = None
    Er: float | None = None

    # TCHEA7
    Ir: float | None = None

    def save(self, path: Path) -> Path:
        """Save model to a JSON file."""
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(self.model_dump_json(indent=2))
        return path

    @classmethod
    def load(cls, path: Path) -> "Composition":
        """Load model from a JSON file."""
        return cls.model_validate_json(path.read_text())

    def elements(self) -> list[str]:
        """Return a list of elements that have non-None values."""
        return [k for k, v in self if k != "name" and v is not None]

    def fractions(self) -> dict[str, float]:
        """Return a dict of element → value (excluding 'name')."""
        return {k: v for k, v in self if k != "name" and v is not None}



from pint import Quantity
from tc_python import (
    CalculationAxis,
    Linear,
    TCPython,
    ThermodynamicQuantity,
    PhaseNameStyle,
)
from typing_extensions import cast

from tc.database.utils import select_thermocalc_database
from tc.schema import Composition, PhaseTransformationTemperatures

T_AXIS_MIN = 500.0
T_AXIS_MAX = 3500.0


def compute_phase_transformation_temperatures(
    composition: Composition,
) -> PhaseTransformationTemperatures:
    """
    Uses Thermo-Calc to compute solidus & liquidus from elements/fractions.
    Returns (solidus_K, liquidus_K, database_name)
    """
    database = select_thermocalc_database(composition)

    elements = composition.elements()
    fractions = composition.fractions()

    with TCPython() as client:
        client.set_cache_folder("cache")
        client.set_ges_version(6)

        calc = (
            client.select_database_and_elements(database, elements)
            .get_system()
            .with_property_diagram_calculation()
            .with_axis(
                CalculationAxis(ThermodynamicQuantity.temperature())
                .set_min(T_AXIS_MIN)
                .set_max(T_AXIS_MAX)
                .with_axis_type(Linear().set_min_nr_of_steps(50))
            )
        )

        # Set composition via mass fractions. Skip the first element; Thermo-Calc will normalize.
        for el, wf in list(fractions.items())[1:]:
            calc.set_condition(f"W({el})", wf)

        diagram = calc.calculate()
        diagram.set_phase_name_style(PhaseNameStyle.ALL)

        groups = diagram.get_values_grouped_by_quantity_of(
            ThermodynamicQuantity.temperature(),
            ThermodynamicQuantity.volume_fraction_of_a_phase("LIQUID"),
        )

    solidus_T = None
    liquidus_T = None

    for group in groups.values():
        xT = group.x
        yL = group.y
        last_zero_idx = max(
            (i for i, y in enumerate(yL) if abs(y) < 1e-12), default=None
        )
        first_one_idx = None
        if last_zero_idx is not None:
            for i in range(last_zero_idx + 1, len(yL)):
                if abs(yL[i] - 1.0) < 1e-12:
                    first_one_idx = i
                    break
        if last_zero_idx is not None:
            solidus_T = xT[last_zero_idx]
        if first_one_idx is not None:
            liquidus_T = xT[first_one_idx]

    if solidus_T is None or liquidus_T is None:
        raise RuntimeError(
            f"Could not determine solidus/liquidus using {database} "
            f"in range [{T_AXIS_MIN:.1f}, {T_AXIS_MAX:.1f}] K."
        )

    temperature_melt = cast(Quantity, Quantity((liquidus_T + solidus_T) / 2, "K"))
    temperature_liquidus = cast(Quantity, Quantity(liquidus_T, "K"))
    temperature_solidus = cast(Quantity, Quantity(solidus_T, "K"))

    phase_transformation_temperatures = PhaseTransformationTemperatures(
        name=composition.name,
        temperature_melt=temperature_melt,
        temperature_liquidus=temperature_liquidus,
        temperature_solidus=temperature_solidus,
    )

    return phase_transformation_temperatures



from enum import Enum


class Alloy(Enum):
    TI6AL4V = "Ti6Al4V"
    IN718 = "IN718"
    SS316L = "SS316L"
    ALSI10MG = "AlSi10Mg"
    IN625 = "IN625"
    MARAGING_STEEL_M300 = "Maraging steel M300"
    TI6AL4V_ELI = "Ti6Al4V ELI"
    SS17_4PH = "SS17-4PH"
    COCRMO = "CoCrMo"
    H13_TOOL_STEEL = "H13 Tool Steel"
    ALSI7MG06 = "AlSi7Mg0.6"
    SS15_5PH = "SS15-5PH"
    CUCRZR = "CuCrZr"
    AL12SI = "Al12Si"
    COPPER = "Copper"
    HASTELLOY_X = "Hastelloy X"
    COCR = "CoCr"
    HASTELLOY_C22 = "Hastelloy C22"
    SS254 = "SS254"
    CUNI2SICR = "CuNi2SiCr"
    D2_TOOL_STEEL = "D2 Tool Steel"
    SS4340 = "SS4340"
    INVAR36 = "Invar36"
    SS304L = "SS304L"
    TI_GRADE_2 = "Ti grade 2"
    K500 = "K500"
    M2_TOOL_STEEL = "M2 Tool Steel"
    TI_GRADE_1 = "Ti grade 1"
    COCR28MO6 = "CoCr28Mo6"
    IN690 = "IN690"
    CUCP = "CuCP"
    TIALNBMO = "TiAlNbMo"
    SS304 = "SS304"
    TIAL = "TiAl"
    HY100 = "HY100"
    ALSI12 = "AlSi12"
    COCRF75 = "CoCrF75"
    AZ91D = "AZ91D"
    COCRW = "CoCrW"
    SS420C = "SS420C"
    SS4130 = "SS4130"
    CUCRZRTI = "CuCrZrTi"
    STEEL_8620L = "8620L Steel"
    AL2319 = "Al2319"
    IRON_0004C = "Iron (0.004%C)"
    SS316 = "SS316"
    IRON_002C = "Iron (0.02%C)"
    MAR_M_247 = "Mar M 247"
    W22FE33NI = "W22Fe33Ni"
    STEEL_4140 = "4140 Steel"
    S50C = "S50C"
    S75C = "S75C"
    A2_TOOL_STEEL = "A2 Tool Steel"
    S105C = "S105C"
    SS_CX = "SS CX"
    ALLOY_7050 = "7050 alloy"
    S33C = "S33C"
    IN600 = "IN600"
    NB = "Nb"



from .__main__ import app

from .alloy_composition import register_alloy_composition
from .alloy_list import register_alloy_list

_ = register_alloy_composition(app)
_ = register_alloy_list(app)

__all__ = ["app"]



from mcp.server import FastMCP


def register_alloy_resources(app: FastMCP):
    from tc.alloy.types import Alloy
    from tc.schema import Composition

    @app.resource("alloy://")
    def alloys() -> list[str] | None:
        from tc.alloy.known_alloy import get_known_alloy_names

        return get_known_alloy_names()

    @app.resource("alloy://{name}/composition")
    def alloy_composition(name: Alloy) -> Composition | None:
        from tc.alloy.known_alloy import get_known_alloy_composition

        return get_known_alloy_composition(name)

    _ = (alloys, alloy_composition)



from mcp.server import FastMCP
from pathlib import Path


def register_phase_transformation_temperatures(app: FastMCP):
    from tc.mcp.types import ToolSuccess, ToolError
    from tc.mcp.utils import tool_success, tool_error
    from tc.schema.composition import Composition

    @app.tool(
        title="Phase Transformation Solidus / Liquidus Temperatures",
        description="Calculate the solidus and liquids temperature given an alloy composition.",
        structured_output=True,
    )
    def phase_transformation_temperatures(
        workspace_name: str,
        composition_filename: str,
    ) -> ToolSuccess[Path] | ToolError:
        from tc.phase_transformation.temperatures import (
            compute_phase_transformation_temperatures,
        )

        from ow.cli.utils import get_workspace_path

        workspace_path = get_workspace_path(workspace_name)

        composition_path = workspace_path / "compositions" / composition_filename
        composition = Composition.load(composition_path)
        phase_transformation_temperatures_path = (
            workspace_path / "phase_transformation_temperatures" / composition.name
        )

        try:
            temperatures = compute_phase_transformation_temperatures(composition)
            temperatures.save(phase_transformation_temperatures_path)
            return tool_success(phase_transformation_temperatures_path)

        except Exception as e:
            return tool_error(
                "Failed to get phase transformation liquidus and solidus temperatures",
                "PHASE_TRANSFORMATION_TEMPERATURES_FAILED",
                exception_type=type(e).__name__,
                exception_message=str(e),
            )

    _ = phase_transformation_temperatures



from mcp.server.fastmcp import FastMCP


def register_schema_composition(app: FastMCP):
    from tc.mcp.types import ToolSuccess, ToolError
    from tc.mcp.utils import tool_success, tool_error
    from tc.schema.composition import Composition

    @app.tool(
        title="Composition Schema",
        description="Creates a composition schema object given a set of elements and their compositions and saves to a given workspace. Inputs to composition should be a fraction (i.e. 95% Fe and 5% C would expect to have inputs of 0.95 Fe and 0.05 C)",
        structured_output=True,
    )
    def schema_composition(
        workspace_name: str,
        name: str,
        Fe: float | None = None,
        C: float | None = None,
        # TCFe alloys
        H: float | None = None,
        Mg: float | None = None,
        Ca: float | None = None,
        Y: float | None = None,
        Ti: float | None = None,
        Zr: float | None = None,
        V: float | None = None,
        Nb: float | None = None,
        Ta: float | None = None,
        Cr: float | None = None,
        Mo: float | None = None,
        W: float | None = None,
        Mn: float | None = None,
        Ru: float | None = None,
        Co: float | None = None,
        Ni: float | None = None,
        Cu: float | None = None,
        Zn: float | None = None,
        B: float | None = None,
        Al: float | None = None,
        Si: float | None = None,
        N: float | None = None,
        P: float | None = None,
        O: float | None = None,
        S: float | None = None,
        Ar: float | None = None,
        Ce: float | None = None,
        # TCNi12 extra elements,
        Hf: float | None = None,
        Re: float | None = None,
        Pd: float | None = None,
        Pt: float | None = None,
        # TCAl9 extra elements
        Li: float | None = None,
        Na: float | None = None,
        K: float | None = None,
        Be: float | None = None,
        Sr: float | None = None,
        Ba: float | None = None,
        Sc: float | None = None,
        Ga: float | None = None,
        Ge: float | None = None,
        Ag: float | None = None,
        In: float | None = None,
        Cd: float | None = None,
        Sn: float | None = None,
        Sb: float | None = None,
        Te: float | None = None,
        Se: float | None = None,
        Pb: float | None = None,
        Bi: float | None = None,
        La: float | None = None,
        Pr: float | None = None,
        Nd: float | None = None,
        Er: float | None = None,
        # TCHEA7
        Ir: float | None = None,
    ) -> ToolSuccess[Composition] | ToolError:
        """
        Creates a configuration file for material properties.
        """

        from ow.cli.utils import get_workspace_path

        try:
            workspace_path = get_workspace_path(workspace_name)

            # Only include arguments that are not None
            values = {
                k: v
                for k, v in locals().items()
                if v is not None and k not in ["workspace_name"]
            }
            composition = Composition(**values)
            composition_path = workspace_path / "compositions" / f"{name}.json"
            composition.save(composition_path)

            return tool_success(composition)

        except Exception as e:
            return tool_error(
                "Failed to create composition schema",
                "SCHEMA_COMPOSITION_FAILED",
                exception_type=type(e).__name__,
                exception_message=str(e),
            )

    _ = schema_composition



from mcp.server.fastmcp import FastMCP

from tc.alloy.mcp import register_alloy_resources, register_alloy_known_alloy
from tc.schema.mcp import register_schema_composition
from tc.phase_transformation.mcp import register_phase_transformation_temperatures

app = FastMCP(name="thermo-calc")

_ = register_alloy_known_alloy(app)
_ = register_alloy_resources(app)
_ = register_schema_composition(app)
_ = register_phase_transformation_temperatures(app)


def main():
    """Entry point for the direct execution server."""
    app.run()


if __name__ == "__main__":
    main()



import typer

from rich import print as rprint


def print_list(name: str, values: list[str] | None = None):
    rprint(f"\n  {name}:")
    if values is None:
        rprint(f"  ⚠️  [yellow]No {name} found.[/yellow]")
    else:
        for index, value in enumerate(values):
            rprint(f"  {index + 1}. [cyan]{value}[/cyan]")


def register_alloy_list(app: typer.Typer):
    @app.command(name="list")
    def alloy_list() -> None:
        """List known alloys."""
        from tc.alloy.known_alloy import get_known_alloy_names

        try:
            names = get_known_alloy_names()
            print_list("Known Alloys", names)
        except:
            rprint("⚠️  [yellow]Unable to list known alloys[/yellow]")
            _ = typer.Exit()

    _ = app.command(name="ls")(alloy_list)

    _ = alloy_list



from typing_extensions import TypedDict
from pintdantic import QuantityDict, QuantityModel, QuantityField

DEFAULT = {
    "name": "Stainless Steel 316L",
    "temperature_melt": (1673, "kelvin"),
    "temperature_liquidus": (1710.26, "kelvin"),
    "temperature_solidus": (1683.68, "kelvin"),
}


class PhaseTransformationTemperaturesDict(TypedDict):
    name: str

    # Melting Temperature (K)
    temperature_melt: QuantityDict

    # Liquidus Temperature (K)
    temperature_liquidus: QuantityDict

    # Solidus Temperature (K)
    temperature_solidus: QuantityDict


class PhaseTransformationTemperatures(QuantityModel):
    """
    Phase Transformation Temperatures calculated by tc-python.
    """

    name: str = DEFAULT["name"]

    temperature_melt: QuantityField = DEFAULT["temperature_melt"]
    temperature_liquidus: QuantityField = DEFAULT["temperature_liquidus"]
    temperature_solidus: QuantityField = DEFAULT["temperature_solidus"]



from mcp.server import FastMCP


def register_alloy_known_alloy(app: FastMCP):
    from tc.mcp.types import ToolSuccess, ToolError
    from tc.mcp.utils import tool_success, tool_error
    from tc.alloy.types import Alloy
    from tc.schema.composition import Composition

    @app.tool(
        title="List Known Alloys",
        description="Provides a list known alloys that have compositions and material properties.",
        structured_output=True,
    )
    def known_alloy_list() -> ToolSuccess[list[str] | None]:
        from tc.alloy.known_alloy import get_known_alloy_names

        known_alloy_names = get_known_alloy_names()

        return tool_success(known_alloy_names)

    @app.tool(
        title="Get Known Alloy Composition",
        description="Provide the elements and composition for a known alloy.",
        structured_output=True,
    )
    def known_alloy_composition(
        workspace_name: str, alloy: Alloy
    ) -> ToolSuccess[Composition | None] | ToolError:
        from tc.alloy.known_alloy import get_known_alloy_composition
        from ow.cli.utils import get_workspace_path

        try:
            workspace_path = get_workspace_path(workspace_name)
            composition = get_known_alloy_composition(alloy)
            composition_path = workspace_path / "compositions" / f"{alloy.name}.json"
            composition.save(composition_path)

            return tool_success(composition)

        except Exception as e:
            return tool_error(
                "Failed to get alloy composition",
                "ALLOY_COMPOSITION_FAILED",
                exception_type=type(e).__name__,
                exception_message=str(e),
            )

    _ = (known_alloy_composition, known_alloy_list)



from .__main__ import app
from .version import register_version

__all__ = ["app"]

from tc.mcp.cli import app as mcp_app
from tc.alloy.cli import app as alloy_app
from tc.schema.cli import app as schema_app
from tc.phase_transformation.cli import app as phase_transformation_app

app.add_typer(mcp_app, name="mcp")
app.add_typer(alloy_app, name="alloy")
app.add_typer(schema_app, name="schema")
app.add_typer(phase_transformation_app, name="phase-transformation")

_ = register_version(app)

if __name__ == "__main__":
    app()



import typer

from rich import print as rprint

from ow.cli.options import WorkspaceOption


def register_phase_transformation_temperatures(app: typer.Typer):
    from tc.schema.composition import Composition

    @app.command(name="temperatures")
    def phase_transformation_temperatures(
        composition_filename: str,
        workspace: WorkspaceOption = None,
    ) -> None:
        """List known alloy composition."""
        from tc.phase_transformation.temperatures import (
            compute_phase_transformation_temperatures,
        )
        from ow.cli.utils import get_workspace_path

        workspace_path = get_workspace_path(workspace)

        composition_path = workspace_path / "compositions" / composition_filename
        composition = Composition.load(composition_path)

        phase_transformation_temperatures_path = (
            workspace_path / "phase_transformation_temperatures" / composition_filename
        )

        try:
            temperatures = compute_phase_transformation_temperatures(composition)
            temperatures.save(phase_transformation_temperatures_path)
            rprint(
                f"✅ [bold green]Phase transformation temperatures saved successfully[/bold green] → {phase_transformation_temperatures_path}"
            )

        except Exception as e:
            rprint("⚠️  [yellow]Unable to determine phase transformation[/yellow]")
            rprint(f"[yellow]Encountered Error: {e}[/yellow]")
            _ = typer.Exit()

    _ = phase_transformation_temperatures



import typer


def register_schema_composition(app: typer.Typer):
    import inspect

    from tc.schema.composition import Composition

    def schema_composition(**kwargs: float | None):
        """Create file for composition."""
        try:
            composition = Composition(**kwargs)
            print(composition.model_dump_json(indent=2))
        except Exception as e:
            typer.secho(f"⚠️ Unable to create composition: {e}", fg=typer.colors.YELLOW)
            raise typer.Exit(code=1)

    # Dynamically build parameters from Composition fields
    params = []
    for field, model_field in Composition.model_fields.items():
        default = model_field.default
        annotation = float | None
        option = inspect.Parameter(
            field,
            inspect.Parameter.KEYWORD_ONLY,
            default=typer.Option(default, help=f"{field} composition fraction"),
            annotation=annotation,
        )
        params.append(option)

    # Assign a new signature to schema_composition
    schema_composition.__signature__ = inspect.Signature(
        parameters=params,
        return_annotation=None,
    )

    app.command("composition")(schema_composition)
    return schema_composition



from tc.schema import Composition


def select_thermocalc_database(composition: Composition) -> str:
    """
    Selects the most appropriate thermo-calc database based on composition
    """
    # Convert the Composition object to a dict and remove None values
    fractions = composition.fractions()

    if not fractions:
        return "PURE5"  # fallback if no elements are set

    # Find the element with the highest fraction
    top_el, top_x = max(fractions.items(), key=lambda kv: kv[1])

    # Count elements with fraction >= 0.15
    multi_principal = sum(1 for x in fractions.values() if x >= 0.15) >= 3

    if top_el == "Ti":
        return "TCTI6"
    if top_el == "Ni" or fractions.get("Ni", 0.0) >= 0.30:
        return "TCNI12"
    if top_el == "Fe":
        return "TCFE14"
    if top_el == "Al":
        return "TCAL9"
    if multi_principal:
        return "TCHEA7"
    return "PURE5"
